import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

/**
 * The class Bird for BitParity implementation.
 * 
 * @author Sjaak Smetsers & Renske Smetsers-Weeda
 * @version 2.0 -- 15-10-2016
 */
public class Bird extends Actor
{
    public static final int NORTH   = 0;
    public static final int EAST    = 1;
    public static final int SOUTH   = 2;
    public static final int WEST    = 3;

    private int myDirection;

    private GreenfootImage imageRight, imageLeft;

    /**
     *  Class constructor indicating which initial direction Bird should face
     *  @param  init_direction which direction Bird should initially be facing
     */
    protected Bird ( int init_direction ) {
        this.myDirection = modulo( init_direction, 4);
        imageRight  = getImage();
        imageLeft = new GreenfootImage( imageRight );
        imageLeft.mirrorHorizontally();
        setImage ();

    }

    private void setImage (){
        if ( myDirection == NORTH ) {
            setImage( imageLeft );
            setRotation( 90 );
        } else if ( myDirection == EAST ) {
            setImage( imageRight );
            setRotation( 0 );
        } else if ( myDirection == SOUTH ) {
            setImage( imageRight );
            setRotation( 90 );
        } else if ( myDirection == WEST ) {
            setImage( imageLeft );
            setRotation( 0 );
        } 
    }

    /**
     * Checks whether there is an egg in the cell ahead
     * 
     * @return true if there is an egg in the cell ahead, false if there is nog egg
     */
    public boolean eggAhead() {
        return getActorAhead( Egg.class ) != null;
    }

    /**
     * Checks whether there is an egg in the current cell
     * 
     * @return true if there is an egg in the current cell, false if there is nog egg
     */
    public boolean foundEgg() {
        return getActor( Egg.class ) != null;
    }

    /**
     * Lays a blue egg in the current cell
     */
    public void layEgg() {
        getWorld().addObject( new BlueEgg (), getX (), getY () );
    }

    /**
     * Removes the egg from the current cell
     */
    public void removeEgg() {
        removeActor( getEgg() );
    }

    /**
     * Gets the reference to the egg-object that is in the current cell
     * 
     * @return reference to the egg-object in the current cell
     */
    public Egg getEgg() {
        return getActor( Egg.class );
    }

    /**
     * Removes the object in the current cell
     */
    private void removeActor( Actor actor ) {
        getWorld().removeObject( actor );
    }

    /**
     * Turns Bird to face in the given direction
     * 
     * @param 0 facing North; 1 facing East; 2 facing South; 3 facing West
     */
    private void setDirection( int new_direction ){
        new_direction = modulo( new_direction, 4 );
        if ( this.myDirection != new_direction ) {
            this.myDirection = new_direction;
            setImage();
            if ( Madagaskar.traceIsOn() ) {
                Greenfoot.delay(1);
            }
        }
    }

    /**
     * Returns a random value between 0 and including 3
     * @return random direction to face
     */
    private int randomDirection() {
        return Greenfoot.getRandomNumber( 4 );
    }

    /**
     * Calculate the modulo
     * @param int a 
     * @param int b
     * @return returns integer result
     */
    private int modulo( int a, int b ) {
        return (a % b + b) % b;
    }

    /**
     * Gets the direction in which the Bird is facing
     * 
     * @return direction in which Bird is facing (0 facing North; 1 facing East; 2 facing South; 3 facing West)
     */
    public int getDirection(){
        return myDirection;
    }

    private <E extends Actor> E getActorAhead(Class<E> cls){
        if ( myDirection == NORTH ) {
            return (E) getOneObjectAtOffset(0,-1, cls);
        } else if ( myDirection == EAST ) {
            return (E) getOneObjectAtOffset(1, 0, cls);
        } else if ( myDirection == SOUTH ) {
            return (E) getOneObjectAtOffset(0, 1, cls);
        } else if ( myDirection == WEST ) {
            return (E) getOneObjectAtOffset(-1,0, cls);
        } else {
            return null;
        }
    }

    private <E extends Actor> E getActor(Class<E> cls){
        return (E) getOneObjectAtOffset(0, 0, cls);
    }

    /**
     * Bird takes one step forward in the direction she is facing
     * 
     */
    public void step() {
        if ( myDirection == NORTH ) {
            setLocation( getX(), getY() - 1 );
        } else if ( myDirection == EAST ) {
            setLocation(getX() + 1, getY() );
        } else if ( myDirection == SOUTH ) {
            setLocation( getX(), getY() + 1 );
        } else if ( myDirection == WEST ) {
            setLocation( getX() - 1, getY() );
        }
        if ( Madagaskar.traceIsOn() ) {
            Greenfoot.delay(1);
        }
    }

    /**
     * Test if there is a border in the cell in front of Bird.
     * 
     * @return true if there is a border in the cell directly in front of Bird, false otherwise
     */
    public boolean borderAhead () {
        if ( myDirection == NORTH ) {
            return getY() == 0;
        } else if ( myDirection == EAST ) {
            return getX() == getWorld().getWidth()  - 1;
        } else if ( myDirection == SOUTH ) {
            return getY() == getWorld().getHeight() - 1;
        } else { // if ( myDirection == WEST ) {
            return getX() == 0;
        }
    }

    /**
     * Test is Bird is facing towards the North
     * 
     * @return true if Bird is facing North, false otherwise
     */
    public boolean facingNorth () {
        return getDirection() == NORTH;
    }

    /**
     * Turns towards the left.
     */
    public void turnLeft() {
        setDirection( modulo( myDirection-1, 4 ) );
    }

    /**
     * Turns towards the right.
     */
    public void turnRight() {
        setDirection( modulo( myDirection+1, 4 ) );
    }

    /**
     * Shows an error popup window with a particular text. Can only be removed by pressing 'Reset'
     * 
     * @param text to display in the popup window
     */
    protected void showError ( String err_msg ) {
        if ( ! Message.messageActive( getWorld() ) ) {
            World my_world = getWorld();
            my_world.addObject (new Alert (err_msg),my_world.getWidth()/2, my_world.getHeight()/2);
        }
    }

    /**
     * Shows an compliment popup window (with a smile icon)s with a particular text. Can only be removed by pressing 'Reset'
     * 
     * @param text to display in the popup window
     */
    protected void showCompliment ( String compl_msg ) {
        if ( ! Message.messageActive( getWorld() ) ) {
            World my_world = getWorld();
            my_world.addObject (new Compliment ( compl_msg ),my_world.getWidth()/2, my_world.getHeight()/2);
        }
    }

    /**
     * Move one cell forward in the current direction, if possible.
     * If not possible, show error popup dialog.
     */
    public void move() {
        if ( canMove() ) {
            step();
        } else {
            showError( "I'm going to fall off the world!" );
        }
    }

    /**
     * Test if we can move forward (thus the end of the world has not been reached).
     * 
     * @return Return true if Dodo can move forward, false otherwise.
     */
    public boolean canMove() {
        if ( ! borderAhead() ){
            return true;
        } else {
            return false;
        }
    } 

}
